@@ -15,6 +15,10 @@
  *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
  */
 
+#ifdef HAVE_CONFIG_H
+#include "config.h"
+#endif
+
 #define _GNU_SOURCE
 #include <stdio.h>
 #include <stdlib.h>
@@ -36,7 +40,8 @@
 #include "settings.h"
 #include "i18n.h"
 
-GdkPixbuf *make_text_bubble(char *text, int *p_width, int *p_height, cowmode_t mode);
+GdkPixbuf *make_text_bubble(char *text, int *p_width, int *p_height,
+                            int max_width, cowmode_t mode);
 GdkPixbuf *make_dream_bubble(const char *file, int *p_width, int *p_height);
 
 #define TICK_TIMEOUT   100
@@ -54,6 +59,7 @@ typedef struct {
    cowstate_t state;
    int transition_timeout;
    int display_time;
+   int screen_width, screen_height;
 } xcowsay_t;
 
 static xcowsay_t xcowsay;
@@ -78,7 +84,6 @@ static cowstate_t next_state(cowstate_t state)
 static GdkPixbuf *load_cow()
 {
    char *cow_path;
-
    const char *alt_image = get_string_option("alt_image");
 
    if (*alt_image)
@@ -197,7 +202,11 @@ static void normal_setup(const char *text, bool debug, cowmode_t mode)
 
    int min_display = get_int_option("min_display_time");
    int max_display = get_int_option("max_display_time");
-   if (xcowsay.display_time < min_display) {
+   if (xcowsay.display_time == 0) {
+      xcowsay.display_time = INT_MAX;
+      debug_msg("Set display time to permanent\n");
+   }
+   else if (xcowsay.display_time < min_display) {
       xcowsay.display_time = min_display;
       debug_msg("Display time too short: clamped to %d\n", min_display);
    }
@@ -205,9 +214,13 @@ static void normal_setup(const char *text, bool debug, cowmode_t mode)
       xcowsay.display_time = max_display;
       debug_msg("Display time too long: clamped to %d\n", max_display);
    }
+
+   const int cow_width = shape_width(xcowsay.cow);
+   const int max_width = xcowsay.screen_width - cow_width;
    
-   xcowsay.bubble_pixbuf = make_text_bubble(text_copy, &xcowsay.bubble_width,
-                                            &xcowsay.bubble_height, mode);
+   xcowsay.bubble_pixbuf = make_text_bubble(
+      text_copy, &xcowsay.bubble_width, &xcowsay.bubble_height,
+      max_width, mode);
    free(text_copy);
 }
 
@@ -225,6 +238,23 @@ static void dream_setup(const char *file, bool debug)
 
 void display_cow(bool debug, const char *text, bool run_main, cowmode_t mode)
 {
+   GdkScreen *screen = gdk_screen_get_default();
+
+   gint n_monitors = gdk_screen_get_n_monitors(screen);
+
+   gint pick = get_int_option("monitor");
+   if (pick < 0 || pick >= n_monitors)
+      pick = random() % n_monitors;
+   
+   GdkRectangle geom;
+   gdk_screen_get_monitor_geometry(screen, pick, &geom);
+
+   xcowsay.screen_width = geom.width;
+   xcowsay.screen_height = geom.height;
+
+   g_assert(xcowsay.cow_pixbuf);
+   xcowsay.cow = make_shape_from_pixbuf(xcowsay.cow_pixbuf);
+
    switch (mode) {
    case COWMODE_NORMAL:
    case COWMODE_THINK:
@@ -238,9 +268,8 @@ void display_cow(bool debug, const char *text, bool run_main, cowmode_t mode)
       exit(1);
    }
    
-   g_assert(xcowsay.cow_pixbuf);
-   xcowsay.cow = make_shape_from_pixbuf(xcowsay.cow_pixbuf);
-
+   xcowsay.bubble = make_shape_from_pixbuf(xcowsay.bubble_pixbuf);   
+   
    int total_width = shape_width(xcowsay.cow)
       + get_int_option("bubble_x")
       + xcowsay.bubble_width;
@@ -248,19 +277,8 @@ void display_cow(bool debug, const char *text, bool run_main, cowmode_t mode)
 
    int bubble_off = max((xcowsay.bubble_height - shape_height(xcowsay.cow))/2, 0);
 
-   GdkScreen *screen = gdk_screen_get_default();
-
-   gint n_monitors = gdk_screen_get_n_monitors(screen);
-
-   gint pick = get_int_option("monitor");
-   if (pick < 0 || pick >= n_monitors)
-      pick = rand() % n_monitors;
-   
-   GdkRectangle geom;
-   gdk_screen_get_monitor_geometry(screen, pick, &geom);
-
-   int area_w = geom.width - total_width;
-   int area_h = geom.height - total_height;
+   int area_w = xcowsay.screen_width - total_width;
+   int area_h = xcowsay.screen_height - total_height;
 
    // Fit the cow on the screen as best as we can
    // The area can't be be zero or we'd get an FPE
@@ -271,29 +289,43 @@ void display_cow(bool debug, const char *text, bool run_main, cowmode_t mode)
 
    int cow_x = get_int_option("at_x");
    if (cow_x < 0)
-      cow_x = rand() % area_w;
+      cow_x = random() % area_w;
    else if (cow_x >= area_w)
       cow_x = area_w - 1;
    
    int cow_y = get_int_option("at_y");
    if (cow_y < 0)
-      cow_y = rand() % area_h;
+      cow_y = random() % area_h;
    else if (cow_y >= area_h)
       cow_y = area_h - 1;
 
-   move_shape(xcowsay.cow,
-      geom.x + cow_x,
-      geom.y + bubble_off + cow_y);
-   show_shape(xcowsay.cow);
-
-   xcowsay.bubble = make_shape_from_pixbuf(xcowsay.bubble_pixbuf);   
-   int bx = shape_x(xcowsay.cow) + shape_width(xcowsay.cow)
-      + get_int_option("bubble_x");
-   int by = shape_y(xcowsay.cow)
-      + (shape_height(xcowsay.cow) - shape_height(xcowsay.bubble))/2
-      + get_int_option("bubble_y");
-   move_shape(xcowsay.bubble, bx, by);
-   
+   if (get_bool_option("left")) {
+      move_shape(xcowsay.cow,
+                 geom.x + cow_x + xcowsay.bubble_width,
+                 geom.y + bubble_off + cow_y);
+      show_shape(xcowsay.cow);
+      
+      int bx = shape_x(xcowsay.cow) - xcowsay.bubble_width
+         + get_int_option("bubble_x");
+      int by = shape_y(xcowsay.cow)
+         + (shape_height(xcowsay.cow) - shape_height(xcowsay.bubble))/2
+         + get_int_option("bubble_y");
+      move_shape(xcowsay.bubble, bx, by);
+   }
+   else {
+      move_shape(xcowsay.cow,
+                 geom.x + cow_x,
+                 geom.y + bubble_off + cow_y);
+      show_shape(xcowsay.cow);
+      
+      int bx = shape_x(xcowsay.cow) + shape_width(xcowsay.cow)
+         + get_int_option("bubble_x");
+      int by = shape_y(xcowsay.cow)
+         + (shape_height(xcowsay.cow) - shape_height(xcowsay.bubble))/2
+         + get_int_option("bubble_y");
+      move_shape(xcowsay.bubble, bx, by);
+   }
+      
    xcowsay.state = csLeadIn;
    xcowsay.transition_timeout = get_int_option("lead_in_time");
    g_timeout_add(TICK_TIMEOUT, tick, NULL);
