@@ -15,6 +15,10 @@
  *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
  */
 
+#ifdef HAVE_CONFIG_H
+#include "config.h"
+#endif
+
 #define _GNU_SOURCE
 #include <stdio.h>
 #include <stdlib.h>
@@ -22,9 +26,9 @@
 #include <getopt.h>
 #include <assert.h>
 #include <string.h>
+#include <limits.h>
 
 #include <sys/types.h>
-#include <sys/stat.h>
 #include <unistd.h>
 #include <errno.h>
 
@@ -34,10 +38,6 @@
 #include "config_file.h"
 #include "i18n.h"
 
-#ifdef HAVE_CONFIG_H
-#include "config.h"
-#endif
-
 // Default settings
 #define DEF_LEAD_IN_TIME  250
 #define DEF_DISPLAY_TIME  CALCULATE_DISPLAY_TIME
@@ -72,6 +72,8 @@ static struct option long_options[] = {
    {"monitor", required_argument, 0, 'm'},
    {"bubble-at", required_argument, 0, 'b'},
    {"at", required_argument, 0, 'a'},
+   {"no-wrap", no_argument, 0, 'w'},
+   {"left", no_argument, 0, 'l'},
    {"config", required_argument, 0, 'o'},
    {"debug", no_argument, &debug, 1},
    {0, 0, 0, 0}
@@ -103,6 +105,7 @@ static void usage()
       " -r, --reading-speed=N\t%s\n"
       " -f, --font=FONT\t%s\n"
       " -d, --dream=FILE\t%s\n"
+      " -l, --left\t\t%s\n"
       "     --think\t\t%s\n"
       "     --daemon\t\t%s\n"
       "     --cow-size=SIZE\t%s\n"
@@ -110,6 +113,7 @@ static void usage()
       "     --monitor=N\t%s\n"
       "     --at=X,Y\t\t%s\n"
       "     --bubble-at=X,Y\t%s\n"
+      "     --no-wrap\t\t%s\n"
       "     --config=FILE\t%s\n"
       "     --debug\t\t%s\n\n"
       "%s\n\n"
@@ -124,6 +128,7 @@ static void usage()
       i18n("Number of milliseconds to delay per word."),
       i18n("Set message font (Pango format)."),
       i18n("Display an image instead of text."),
+      i18n("Make the bubble appear to the left of cow."),
       i18n("Display a thought bubble rather than a speech bubble."),
       i18n("Run xcowsay in daemon mode."),
       i18n("Size of the cow (small, med, large)."),
@@ -131,6 +136,7 @@ static void usage()
       i18n("Display cow on monitor N."),
       i18n("Force the cow to appear at screen location (X,Y)."),
       i18n("Change relative position of bubble."),
+      i18n("Disable wrapping if text cannot fit on screen."),
       i18n("Specify alternative config file."),
       i18n("Keep daemon attached to terminal."),
       i18n("Default values for these options can be specified in the "
@@ -240,11 +246,13 @@ int main(int argc, char **argv)
    add_int_option("bubble_x", DEF_BUBBLE_X);
    add_int_option("bubble_y", 0);
    add_string_option("alt_config_file", "");
+   add_bool_option("wrap", true);
+   add_bool_option("left", false);
    
    parse_config_file();
    
    int c, index = 0, failure = 0;
-   const char *spec = "hvd:rt:f:";
+   const char *spec = "hvld:rt:f:";
    const char *dream_file = NULL;
    while ((c = getopt_long(argc, argv, spec, long_options, &index)) != -1) {
       switch (c) {
@@ -298,6 +306,12 @@ int main(int argc, char **argv)
          set_string_option("alt_config_file", optarg);
          parse_config_file();
          break;
+      case 'w':
+         set_bool_option("wrap", false);
+         break;
+      case 'l':
+         set_bool_option("left", true);
+         break;
       case '?':
          // getopt_long already printed an error message
          failure = 1;
@@ -309,7 +323,7 @@ int main(int argc, char **argv)
    if (failure)
       exit(EXIT_FAILURE);
 
-   srand((unsigned)time(NULL));
+   srandom((unsigned)time(NULL));
 
    cowmode_t mode = think_flag ? COWMODE_THINK : COWMODE_NORMAL;
 
@@ -321,17 +335,14 @@ int main(int argc, char **argv)
 
       if (dream_file != NULL) {
          // Make path absolute
-         char *wd = getcwd(NULL, 0);
-         char *abs_path;
+         char *abs_path = realpath(dream_file, NULL);
+         if (abs_path == NULL) {
+            perror(dream_file);
+            exit(EXIT_FAILURE);
+         }
 
-         asprintf(&abs_path, "%s/%s", wd, dream_file);
-         free(wd);
-
-         struct stat dummy;
-         if (stat(abs_path, &dummy) != 0) {
-            fprintf(stderr, "Error: %s: %s\n", abs_path,
-               strerror(errno));
-            free(abs_path);
+         if (access(abs_path, R_OK) != 0) {
+            perror(abs_path);
             exit(EXIT_FAILURE);
          }
          
